#!/usr/bin/env python
# coding: utf-8

"""
Tools for the generation and analysis of dislocation distributions.

A dislocation is characterized by a position and a Burgers vector. A
distribution is a set of dislocations in a material. The dislocations
of a distribution are generated randomly according to a model. A sample
is a set of distributions generated with the same model.

A distribution can be exported as a map showing the location of its
dislocations. It can also be exported in a standardized form for input
to an X-ray diffraction simulation program. Finally it is possible to
make a spatial analysis of a model by means of statistical functions
on a sample of distributions.

In this package, Burgers vectors are represented only by their sense
(positive: +1, or negative: -1). Their direction is only determined in
the header of the exported input data files. It is parallel to the
dislocation line for the 'screw' type or perpendicular for the 'edge'
type. The superposition principle simplifies the analysis by allowing
to study only one direction at a time.

To obtain relevant statistical results, the spatial analysis must be
averaged over a large number of distributions. To reduce the
computation time it is possible to parallelize the calculations. The
MPI library can be used for this purpose. A sample of distributions is
generated and analyzed on each core, then the results are averaged
over the cores.
"""

__author__ = "Dunstan Becht"
__version__ = "0.9.3"

import sys
import numpy as np
from typing import Union, Optional, NewType, Any
from beartype import beartype

if sys.version_info[0]>=3 and sys.version_info[1]>=9:
    from collections.abc import Callable
    List = list
    Tuple = tuple
else:
    from typing import Callable
    from typing import List
    from typing import Tuple

Scalar = Union[int, float, np.intc]
Vector = np.ndarray # shape: (n,)
ScalarList = np.ndarray # shape: (...,)
ScalarListList = np.ndarray # shape: (..., ...)
VectorList = np.ndarray # shape: (..., n)
CorrectionFunction = Callable[[Vector, ScalarList, ScalarList], ScalarList]
GenerationFunction = Callable[[str, int], Tuple[VectorList, VectorList]]
AnalysisOutput = Union[Tuple, Union[Scalar, np.ndarray]]



#!/usr/bin/env python
# coding: utf-8

"""
Tools for spatial analysis of dislocation distributions.
"""

import matplotlib.pyplot as plt
from . import *
from . import sets

@beartype
def N(
    a: Vector,
    B: VectorList,
    r2: ScalarList,
) -> ScalarList:
    """
    Return the number of points of B in the neighborhoods of a.

    The function does not use the notion of dislocation. It allows only
    a spatial analysis of a distribution of points in space. If a point
    of B coincides with the point a then it is not counted.

    Input:
        a: point around which the neighborhoods are formed
        B: observed points that are counted when in the neighborhoods
        r2: squared radius of the neighborhoods in ascending order

    Output:
        n: list of the number of points for each radius value

    Complexity:
        O( len(B)*log(len(B)) + len(r2) )
    """
    d2 = np.sum(np.square(np.subtract(B, a)), axis=1) # squared distances to a
    d2 = np.sort(d2[(d2<=r2[-1]) & (d2>0)]) # sorted squared distances to a
    j, dn = 0, np.zeros(len(r2), dtype=np.int64)
    for i in range(len(d2)):
        while d2[i]>r2[j]:
            j += 1
        dn[j] += 1
    return np.cumsum(dn)

@beartype
def M(
    A: VectorList,
    B: VectorList,
    w: CorrectionFunction,
    r: ScalarList,
    r2: ScalarList,
) -> ScalarList:
    """
    Return the value of N corrrected and averaged on the points of A.

    The function does not use the notion of dislocation. It allows only
    a spatial analysis of a distribution of points in space.

    Input:
        A: points around which the neighborhoods are formed
        B: observed points that are counted when in the neighborhoods
        w: weighting function for correction at the edges
        r: radius of the neighborhoods in ascending order
        r2: squared radius of the neighborhoods in ascending order

    Output:
        m: list of the average number of points for each radius value

    Complexity:
        O( len(A) * (complexity(w)+complexity(N)) )
    """
    s = np.zeros(len(r))
    for i in range(len(A)):
        s += w(A[i], r, r2) * N(A[i], B, r2)
    return s/len(A)

@beartype
def MMMM_cp_cm(
    d: sets.Distribution,
    r: ScalarList,
    r2: ScalarList,
) -> Tuple[ScalarListList, Scalar, Scalar]:
    """
    Return M++, M-+, M+-, M-- and the number of + and - dislocations.

    M++ is the average number of dislocations with positive Burgers
    vector sense around dislocations with positive sense. The
    dislocation in the center of the neighborhood is not counted.

    M-+ is the average number of dislocations with positive Burgers
    vector sense around dislocations with negative sense.

    M+- is the average number of dislocations with negative Burgers
    vector sense around dislocations with positive sense.

    M-- is the average number of dislocations with negative Burgers
    vector sense around dislocations with negative sense. The
    dislocation in the center of the neighborhood is not counted.

    Input:
        d: distribution of dislocations to analyze
        r: radius of the neighborhoods in ascending order [nm]
        r2: squared radius of the neighborhoods [nm^2]

    Output:
        MMMM: stacked values of M++, M-+, M+-, M-- [1]
        cp: number of dislocations with Burgers vector sense + [1]
        cm: number of dislocations with Burgers vector sense - [1]

    Complexity:
        O( complexity(M) )
    """
    Pp = d.p[d.b>0]
    Pm = d.p[d.b<0]
    Mpp = M(Pp, Pp, d.w, r, r2)
    Mmp = M(Pm, Pp, d.w, r, r2)
    Mpm = M(Pp, Pm, d.w, r, r2)
    Mmm = M(Pm, Pm, d.w, r, r2)
    return np.stack((Mpp, Mmp, Mpm, Mmm)), len(Pp), len(Pm)

@beartype
def KKKK_dp_dm(
    MMMM: ScalarListList,
    cp: Scalar,
    cm: Scalar,
    v: Scalar,
) -> Tuple[ScalarListList, Scalar, Scalar]:
    """
    Return K++, K-+, K+-, K-- and densities of + and - dislocations.

    Implementation of the Ripley’s K function.

    This function is incremental. It uses pre-calculated MMMM values
    to avoid unnecessary repetition of calculations common to spatial
    analysis functions

    Input:
        MMMM: stacked values of M++, M-+, M+-, M-- [1]
        cp: number of dislocations with Burgers vector sense + [1]
        cm: number of dislocations with Burgers vector sense - [1]
        v: n-volume of the region of interest [nm^n]

    Output:
        KKKK: stacked values of K++, K-+, K+-, K-- [nm^n]
        dp: density of dislocations with Burgers vector sense + [nm^-n]
        dm: density of dislocations with Burgers vector sense - [nm^-n]

    Complexity:
        O( len(MMMM) )
    """
    dp, dm = cp/v, cm/v
    return np.concatenate((MMMM[0:2]/dp, MMMM[2:4]/dm)), dp, dm

@beartype
def gggg_dVvdr(
    KKKK: ScalarListList,
    r: ScalarList,
    n: int,
) -> Tuple[ScalarListList, ScalarList]:
    """
    Return g++, g-+, g+-, g-- and derivative of neighborhood volume.

    This function is incremental. It uses pre-calculated KKKK values
    to avoid unnecessary repetition of calculations common to spatial
    analysis functions

    Input:
        KKKK: stacked values of K++, K-+, K+-, K-- [nm^n]
        r: radius of the neighborhoods in ascending order [nm]
        n: dimension of the space of the region of interest

    Ouput:
        gggg: stacked values of g++, g-+, g+-, g-- [1]
        dVvdr: derivative of the neighborhood volume [nm^(n-1)]

    Complexity:
        O( len(r) )
    """
    if n == 2:
        dVvdr = np.where(r!=0, 2*np.pi*r, 1)
    elif n == 3:
        dVvdr = np.where(r!=0, np.pi*r**2, 1)
    dKKKK = np.roll(KKKK, -1) - np.roll(KKKK, 1)
    dr = np.roll(r, -1) - np.roll(r, 1)
    dKdr = np.divide(dKKKK, dr)
    gggg = dKdr/dVvdr
    for i in range(len(gggg)):
        gggg[i][0], gggg[i][-1] = None, None
    return gggg, dVvdr

@beartype
def GaGs(
    MMMM: ScalarListList,
    cp: Scalar,
    cm: Scalar,
) -> ScalarListList:
    """
    Return Ga and Gs.

    This function is incremental. It uses pre-calculated MMMM values
    to avoid unnecessary repetition of calculations common to spatial
    analysis functions

    Input:
        MMMM: stacked values of M++, M-+, M+-, M-- [1]
        cp: number of dislocations with Burgers vector sense + [1]
        cm: number of dislocations with Burgers vector sense - [1]

    Output:
        GaGs: stacked values of Ga and Gs [1]

    Complexity:
        O( len(MMMM) )
    """
    Ga = cp*(MMMM[0]-MMMM[2]) + cm*(MMMM[3]-MMMM[1])
    Gs = cp*(MMMM[0]+MMMM[2]) + cm*(MMMM[3]+MMMM[1])
    return np.stack((Ga, Gs))

@beartype
def calculate(
    q: List[str],
    o: Union[sets.Distribution, sets.Sample],
    r: ScalarList,
    r2: Optional[ScalarList] = None,
) -> AnalysisOutput:
    """
    Return the values of the quantities in q calculated incrementally.

    The following quantities can be requested:
        'MMMM': stacked values of M++, M-+, M+-, M-- [1]
        'KKKK': stacked values of K++, K-+, K+-, K-- [nm^n]
        'gggg': stacked values of g++, g-+, g+-, g-- [1]
        'GaGs': stacked values of Ga and Gs [1]
        'cp': number of dislocations with Burgers vector sense + [1]
        'cm': number of dislocations with Burgers vector sense - [1]
        'dp': density of dislocations with Burgers vector sense + [nm^-n]
        'dm': density of dislocations with Burgers vector sense - [nm^-n]
        'dVvdr': derivative of the neighborhood volume [nm^-(n-1)]

    Input:
        q: name of the quantities to calculate
        o: distribution or sample of distributions to analyze
        r: radius of the neighborhoods in ascending order [nm]
        r2: squared radius of the neighborhoods in ascending order [nm^2]

    Output:
        v: values of the quantities in the order requested in q

    Exemple:
        >>> KKKK, GaGs = calculate(['KKKK', 'GaGs'], distribution, r)

    Complexity:
        O( complexity(MMMM_Pp_Pm) ) if o is a distribution
        O( complexity(MMMM_Pp_Pm) * len(o) ) if o is a sample
    """
    if r2 is None:
        r2 = np.square(r)

    @beartype
    def calculate_on_distribution(
        d: sets.Distribution,
    ) -> AnalysisOutput:
        """
        Auxiliary function of calculate applied to a distribution only.
        """
        s = {}
        s['MMMM'], s['cp'], s['cm'] = MMMM_cp_cm(
            d,
            r,
            r2,
        )
        if 'gggg' in q or 'KKKK' in q:
            s['KKKK'], s['dp'], s['dm'] = KKKK_dp_dm(
                s['MMMM'],
                s['cp'],
                s['cm'],
                d.v,
            )
        if 'gggg' in q:
            s['gggg'], s['dVvdr'] =  gggg_dVvdr(
                s['KKKK'],
                r,
                d.n,
            )
        if 'GaGs' in q:
            s['GaGs'] = GaGs(
                s['MMMM'],
                s['cp'],
                s['cm'],
            )
        return tuple([s[k] for k in q])

    if isinstance(o, sets.Distribution):
        return calculate_on_distribution(o)
    return o.average(calculate_on_distribution)

@beartype
def plot_KKKK(
    r: ScalarList,
    KKKK: ScalarListList,
    p: str,
    n: str,
    t: str,
) -> None:
    """
    Export the figure showing functions K++, K-+, K+-, K--.

    Input:
        r: radius of the neighborhoods [nm]
        KKKK: stacked values of K++, K-+, K+-, K-- [nm^n]
        p: path where to export the file
        n: name of the exported file
        t: title of the plot

    Complexity:
        O( len(r) )
    """
    # K of a uniform distribution
    r_compare = np.linspace(r[0], r[-1], 10)
    k_compare = np.pi*r_compare**2
    # fig
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6))
    fig.subplots_adjust(left=0.06, right=0.98, bottom=0.1)
    fig.suptitle(r"$ K_{ab} $ of "+t, fontsize=16)
    # ax1
    ax1.plot(r_compare, k_compare, "*", label=r"$ \pi r^2 $")
    ax1.plot(r, KKKK[0], label=r"$K_{++}(r)$")
    ax1.plot(r, KKKK[1], label=r"$K_{-+}(r)$")
    ax1.legend()
    ax1.grid()
    ax1.set_xlabel(r"$r \ (nm)$")
    # ax2
    ax2.plot(r_compare, k_compare, "*", label=r"$ \pi r^2 $")
    ax2.plot(r, KKKK[3], label=r"$K_{--}(r)$")
    ax2.plot(r, KKKK[2], label=r"$K_{+-}(r)$")
    ax2.legend()
    ax2.grid()
    ax2.set_xlabel(r"$r \ (nm)$")
    ax2.set_ylabel(r"$(nm^2)$")
    # export
    plt.savefig(p+"KKKK_"+n+".pdf")
    plt.close('all')

@beartype
def plot_gggg(
    r: ScalarList,
    gggg: ScalarListList,
    p: str,
    n: str,
    t: str,
) -> None:
    """
    Export the figure showing functions g++, g-+, g+-, g--.

    Input:
        r: radius of the neighborhoods [nm]
        gggg: stacked values of g++, g-+, g+-, g-- [1]
        p: path where to export the file
        n: name of the exported file
        t: title of the plot

    Complexity:
        O( len(r) )
    """
    # fig
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6))
    fig.subplots_adjust(left=0.06, right=0.98, bottom=0.1)
    fig.suptitle(r"$ g_{ab} $ of "+t, fontsize=16)
    ymin = min(np.nanmin(gggg)*0.95, 0.9)
    ymax = max(np.nanmax(gggg)*1.05, 1.1)
    # ax1
    ax1.plot(r, gggg[0], label=r"$g_{++}(r)$")
    ax1.plot(r, gggg[1], label=r"$g_{-+}(r)$")
    ax1.legend()
    ax1.grid()
    ax1.set_xlabel(r"$r \ (nm)$")
    ax1.set_ylim(ymin, ymax)
    # ax2
    ax2.plot(r, gggg[3], label=r"$g_{--}(r)$")
    ax2.plot(r, gggg[2], label=r"$g_{+-}(r)$")
    ax2.legend()
    ax2.grid()
    ax2.set_xlabel(r"$r \ (nm)$")
    ax2.set_ylim(ymin, ymax)
    # export
    plt.savefig(p+"gggg_"+n+".pdf")
    plt.close('all')

@beartype
def plot_GaGs(
    r: ScalarList,
    GaGs: ScalarListList,
    p: str,
    n: str,
    t: str,
) -> None:
    """
    Export the figure showing functions Ga and Gs.

    Input:
        r: radius of the neighborhoods [nm]
        GaGs: stacked values of Ga and Gs [1]
        p: path where to export the file
        n: name of the exported file
        t: title of the plot

    Complexity:
        O( len(r) )
    """
    # fig
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 6))
    fig.subplots_adjust(left=0.06, right=0.98, bottom=0.1)
    fig.suptitle(r"$ G^a $ and $ G^s $ of "+t, fontsize=16)
    # ax1
    ax1.plot(r, GaGs[0], label=r"$G^a(r)$")
    ax1.legend()
    ax1.grid()
    ax1.set_xlabel(r"$r \ (nm)$")
    # ax2
    ax2.plot(r, GaGs[1], label=r"$G^s(r)$")
    ax2.legend()
    ax2.grid()
    ax2.set_xlabel(r"$r \ (nm)$")
    # export
    plt.savefig(p+"GaGs_"+n+".pdf")
    plt.close('all')

@beartype
def intervals(
    i: Scalar,
    s: Scalar,
) -> Tuple:
    """
    Suggest default radius interval for the analysis of a distribution.

    It is proposed to conduct the analysis for a radius varying from 0
    to half the characteristic size of the region of interest. For the
    K-function plot it is proposed to vary the radius from 0 to 4 times
    the inter-dislocation distance.

    Input:
        i: inter dislocation distance [nm]
        s: shape size [nm]

    Output:
        r: full range of radii for analysis [nm]
        iK: index of the maximum value for the plot of K
    """
    a, b, c = 0, 4*i, s/2
    n = 100 # number of points between a and b
    return np.linspace(a, c, int((c-a)*n/(b-a))), n

@beartype
def export(
    o: Union[sets.Distribution, sets.Sample],
    p: str = "",
    n: Optional[str] = None,
    t: Optional[str] = None,
) -> None:
    """
    Export a complete analysis of the object o.

    Input:
        o: distribution or sample of distributions to analyze
        p: path where to export the file
        n: name of the exported file
        t: title of the plot

    Complexity:
        O( complexity(calculate) )
    """
    if p!="" and p[-1]!="/":
        p += "/"
    r, iK = intervals(o.i, o.s) # intervals to display
    KKKK, gggg, GaGs = calculate(['KKKK', 'gggg', 'GaGs'], o, r)
    n = o.fileName(t=False)
    t = o.plotTitle()
    plot_KKKK(r[:iK], KKKK.T[:iK].T, p, n, t)
    plot_gggg(r, gggg, p, n, t)
    plot_GaGs(r, GaGs, p, n, t)



#!/usr/bin/env python
# coding: utf-8

"""
Tools to standardize notations.
"""

from . import *

@beartype
def number(
    x: Scalar,
    t: str = 'file',
    w: int = 4,
) -> str:
    """
    Return a notation of x that can be used in a file name.

    Input:
        x: number
        t: type of use ('file', 'console', 'title')
        w: width of the output string

    Output:
        s: string of x
    """
    if x<10**w and x>=1 or x == 0: # diplay x with fixed comma
        s = str(round(x))
        if t == 'file':
            s = s.zfill(w) # fill with zeros
        elif t == 'console':
            s = format(s, '>'+str(w)) # fill with spaces
    elif t == 'title': # LaTeX scientific notation
        s = format(x, '1.'+str(max(w-6, 0))+'e')
        m, e = s.split("e") # separate mantissa and exponent
        s = m+" \\times 10^{"+str(int(e))+"}"
    else: # scientific notation with 'e'
        s = format(x, '1.0e').replace("+", "")
        s = s.replace("e0", "e").replace("e-0", "e-")
        d = w-len(s) # number of missing characters to reach the ideal width
        if d>0:
            if t=='file' or d==1:
                if "e-" in s:
                    s = s.replace("e-", "e-0")
                else:
                    s = s.replace("e", "e0")
            else:
                fmt = '1.'+str(d-1)+'e'
                s = format(x, fmt)
                s = s.replace("+", "").replace("e0", "e").replace("e-0", "e-")
    return s

@beartype
def parameters(
    r: dict,
    t: str = 'file',
) -> str:
    """
    Return a string descibing the model parameters.

    Input:
        r: model parameters
        t: type of use ('file', 'console', 'title')

    Output:
        n: model parameters description
    """
    # user-defined name for the set of parameters
    if 'name' in r:
        return r['name']
    n = ""
    # writing of equalities
    if t == 'file':
        weq = lambda a, b: a+b
    elif t == 'console':
        weq = lambda a, b: a+"="+b
    elif t == 'title':
        weq = lambda a, b: a+r" = "+b
    # version
    if 'v' in r:
        n += "-"+r['v']
    # parameter list
    p = []
    if 'd' in r:
        p.append(weq('d', number(r['d'], t))) # density
    if 's' in r:
        p.append(weq("s", number(r['s'], t, 4))) # subarea or cell side
    if 'f' in r:
        p.append(weq("f", str(r['f']))) # filling
    if 't' in r:
        p.append(weq("t", number(r['t'], t, 3))) # wall thickness
    if 'l' in r:
        p.append(weq("l", number(r['l'], t, 3))) # dipole length
    if 'r' in r:
        p.append(weq("r", str(r['r'])))
    # concatenate
    if len(p) > 0:
        if t == 'title':
            n += r" $ \left( "+", \ ".join(p)+r" \right) $ "
        elif t == 'console':
            n += " ("+" ".join(p)+")"
        else:
            n += "_"+"_".join(p)
    return n



#!/usr/bin/env python
# coding: utf-8

"""
Models for the random generation of dislocations.

The following parameters can be used in the model parameters:
    r: random seed (RDD, RRDD, RCDD)
    v: variant (RRDD, RCDD)
    d: density (RDD, RRDD, RCDD)
    s: side of the subarea (RRDD) or the cell (RCDD)
    f: filling of the subareas (RRDD)
    l: dipole length (RCDD)
    name: parameters set nickname (RDD, RRDD, RCDD)
"""

import math
from . import *

@beartype
def seed(
    r: dict,
) -> None:
    """
    Set the specified seed or randomly select a seed if not specified.

    Letting the user choose a seed allows the replicability of the
    generation of a distribution. This function must be called at the
    start of the model definitions.

    Input:
        r: model parameters

    Complexity:
        O( 1 )
    """
    if not 'r' in r:
        r['r'] = np.random.randint(0, 2**32, dtype=np.int64)
    np.random.seed(r['r'])

@beartype
def RDD(
    g: str,
    s: Scalar,
    a: Scalar,
    r: dict,
) -> Tuple[VectorList, ScalarList]:
    """
    Return the positions and Burgers vector generated with RDD model.

    "RDD" means: "Random Dislocation Distribution"

    Input:
        g: geometry of the region of interest
        s: size of the region of interest [nm]
        a: area of region of interest [nm^2]
        r: model parameters

    Output:
        p: dislocation positions [nm]
        b: dislocation Burgers vectors sense [1]

    The following parameters can be specified in r:
        'r' (int): random seed
        'd' (Scalar): density of dislocations [nm^-2]

    When the parameter 'r' is not specified, the random seed is chosen
    randomly and added to the parameters dictionary.

    Complexity:
        O( r['d'] * a )
    """
    # parameters
    seed(r) # random seed
    rd = r['d'] # density of dislocations
    n = round(rd*a) # number of dislocations
    # dislocation positions
    if g == 'circle':
        theta = 2*np.pi*np.random.random(n)
        radius = s*np.sqrt(np.random.random(n))
        p = np.stack((radius*np.cos(theta), radius*np.sin(theta)), axis=1)
    elif g == 'square':
        p = s*np.random.random([n, 2])
    # dislocation Burgers vectors
    bp = np.ones(n//2, dtype=int) # positive Burgers vectors
    b = np.concatenate((bp, -bp)) # add negative Burgers vectors
    if n%2 == 1: # add a random burger vector
        b = np.concatenate((b, np.random.choice([1, -1], size=1)))
    return p, b

@beartype
def ticks(
    g: str,
    l: Scalar,
    s: Scalar,
) -> ScalarList:
    """
    Standardize the construction of a grid of subareas or cells.

    Input:
        g: geometry
        l: length to subdivide [nm]
        s: size of the step [nm]

    Output:
        t: ticks along an axis

    Complexity:
        O( l/s )
    """
    m = math.ceil(l/s)
    if g == 'circle':
        return s*np.arange(-m, m+1) # [-2s, -s, 0, s, 2s] with m=2
    if g == 'square':
        return s*np.arange(m+1) # [0, s, 2s] with m=2

@beartype
def even_positions(
    t: VectorList,
    f: int,
) -> Tuple[ScalarList, ScalarList]:
    """
    Return the positions with a number of points f in each case.

    Input:
        t: ticks along an axis
        f: number of points in each case

    Output:
        x: x coordinates of points
        y: y coordinates of points

    Example for t=[0, 1] and f=4:
        x = [0 0 0 0 1 1 1 1 0 0 0 0 1 1 1 1]
        y = [0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1]
    """
    x = np.repeat(np.tile(t, len(t)), f)
    y = np.repeat(t, f*len(t))
    return x, y

@beartype
def even_senses(
    t: VectorList,
    f: int,
) -> ScalarList:
    """
    Return the evenly distributed Burgers senses for even_positions.

    Input:
        t: ticks
        f: number of points in each case

    Output:
        b: Burgers vector senses

    Example for t=[0, 1] and f=4:
        b = [- - + + - - + + - - + + - - + +]
    """
    bp = np.ones(f//2, dtype=int)
    b = np.tile(np.concatenate((bp, -bp)), len(t)**2)
    return b

@beartype
def RRDD(
    g: str,
    s: Scalar,
    a: Scalar,
    r: dict,
) -> Tuple[VectorList, ScalarList]:
    """
    Return the positions and Burgers vector generated with RRDD model.

    "RRDD" means: "Restrictedly Random Dislocation Distribution"

    Input:
        g: geometry of the region of interest
        s: size of the region of interest [nm]
        a: area region of interest [nm^2]
        r: model parameters

    Output:
        p: dislocation positions [nm]
        b: dislocation Burgers vectors sense [1]

    The following parameters can be specified in r:
        'r': (int): random seed
        'v' (str): variant ('R' or 'E')
        's' (int): side of the subareas [nm]
        'f' (int): number of dislocations per subarea
        'd' (Scalar): density of dislocations [nm^-2]

    When the parameter 'r' is not specified, the random seed is chosen
    randomly and added to the parameters dictionary. The parameter 'd'
    is considered only if the parameter 'f' is not specified.

    When the parameter 'v' is set to 'R', the Burgers vectors are
    randomly distributed. When it is set to 'E', the Burgers vectors
    are evenly distributed in each subarea.

    Complexity:
        O( r['f'] * s^2/r['s']^2 )
    """
    # parameters
    seed(r) # random seed
    rv = r['v'] # model variant
    rs = r['s'] # size of the subareas
    if 'f' in r:
        rf = r['f'] # specified filling
    else:
        rf = round(r['d']*rs**2) # deduced filling
    if rf%2 != 0:
        raise ValueError('odd number of dislocations per subarea')
    t = ticks(g, s, rs)[:-1] # left or bottom location of subareas
    n = rf * len(t)**2 # number of dislocations
    # dislocation positions
    x, y = even_positions(t, rf)
    p = np.stack((x,y), axis=1) + rs*np.random.random((n, 2))
    # dislocation Burgers vector
    if rv == 'R':
        b = np.random.choice([1, -1], size=n)
    elif rv == 'E':
        b = even_senses(t, rf)
    # masking
    if g == 'circle':
        mask = np.sum(np.square(p), axis=1) < s**2
    elif g == 'square':
        mask = (p[:,0]<s) & (p[:,1]<s)
    return p[mask], b[mask]

@beartype
def RCDD(
    g: str,
    s: Scalar,
    a: Scalar,
    r: dict,
) -> Tuple[VectorList, ScalarList]:
    """
    Return the positions and Burgers vector generated with RCDD model.

    "RCDD" means: "Random Cell Dislocation Distribution"

    Input:
        g: geometry of the region of interest
        s: size of the region of interest [nm]
        a: area region of interest [nm^2]
        r: model parameters

    Output:
        p: dislocation positions [nm]
        b: dislocation Burgers vectors sense [1]

    The following parameters can be specified in r:
        'r': (int): random seed
        'v' (str): variant ('R', 'E' or 'D')
        'd' (Scalar): density of dislocations [nm^-2]
        's' (int): side of the subareas [nm]
        't' (int): thickness of the cell walls [nm]
        'l' (int): length of the dipoles [nm]

    When the parameter 'r' is not specified, the random seed is chosen
    randomly and added to the parameters dictionary.

    When the parameter 'v' is set to 'R', the Burgers vectors are
    randomly distributed. When 'v' is set to 'E', the Burgers vectors
    are evenly distributed in each cell. When 'v' is set to 'D', the
    Burgers vectors with positive sense and negative sense are spaced
    by a distance defined by 'l'.

    Complexity:
        O( r['d'] * a )
    """
    # parameters
    seed(r)
    rv = r['v'] # variant
    rd = r['d'] # density of dislocations
    rs = r['s'] # side of the cells
    rt = r['t'] # thickness of the cell walls
    if rt > rs/2:
        raise ValueError("inconsistent thickness and cell side")
    if rv == 'D':
        rl = r['l'] # length of the dipoles
    t = ticks(g, s, rs)[:-1] # left or bottom location of subareas
    n = round(rd*rs**2*len(t)**2) # number of dislocations
    # dislocation or dipole positions
    if rv == 'D':
        c = n//2 # 1 random point for 1 dipole of 2 dislocations
    elif rv == 'R':
        c = n # 1 random point for 1 dislocation
    elif rv == 'E':
        rf = round(r['d']*rs**2) # deduced filling
        if rf%2 != 0:
            raise ValueError("odd number of dislocations per cell")
        c = rf * len(t)**2 # 1 random point for 1 dislocation
    # choice of cells
    if rv in ['R', 'D']:
        x = 1.0*np.random.choice(t, size=c)
        y = 1.0*np.random.choice(t, size=c)
    elif rv == 'E':
        x, y = even_positions(t, rf)
    # choice of position in the cells
    l1 = rt/2
    l2 = rs - l1
    m = np.random.randint(4, size=c)
    rx = np.random.random(c)
    ry = np.random.random(c)
    np.add(x, l1 + rx*l2, x, where=m==0)
    np.add(y,      ry*l1, y, where=m==0)
    np.add(x,      rx*l2, x, where=m==1)
    np.add(y, l2 + ry*l1, y, where=m==1)
    np.add(x,      rx*l1, x, where=m==2)
    np.add(y,      ry*l2, y, where=m==2)
    np.add(x, l2 + rx*l1, x, where=m==3)
    np.add(y, l1 + ry*l2, y, where=m==3)
    # Burgers vector
    if rv == 'R':
        b = np.random.choice([1, -1], size=c)
    elif rv == 'E':
        b = even_senses(t, rf)
    elif rv == 'D':
        b = np.concatenate((np.ones(c), -np.ones(c)))
        theta = 2*np.pi*np.random.random(c)
        ux = rl/2 * np.cos(theta)
        uy = rl/2 * np.sin(theta)
        x = np.concatenate((x+ux, x-ux))
        y = np.concatenate((y+uy, y-uy))
    p = np.stack((x,y), axis=1)
    # masking
    if g == 'circle':
        mask = np.sum(np.square(p), axis=1) < s**2
        p = p[mask]
        b = b[mask]
    return p, b



#!/usr/bin/env python
# coding: utf-8

"""
Tools for representing distributions and samples of distributions.
"""

from . import *
from . import models
from . import overlap
from . import notation
from . import boundaries

class Distribution:
    """
    Represent a distribution of dislocations.

    When the geometry is of type 'circle' the characteristic size is
    its radius. When it is of type 'square' the characteristic size is
    its side.

    The dislocations can be of type 'screw' or type 'edge'. If the type
    is not specified, type 'screw' is chosen by default.

    For circle geometry, if the parameter c is set to 'IDBC', image
    dislocations are added to the distribution.

    For square geometry, the parameter c can be set to 'PBCG<r>' or
    'PBCR<r>' where <r> is the number of replicates of the region of
    interest around the boundaries. With 'PBCG<r>' the replicated
    dislocations are added to the distribution. With 'PBCR<r>' the
    replicated dislocations are not added to the distribution but the
    X-ray diffraction simulation program is warned to replicate the
    region of interest.

    Attributes:
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        n (int): dimension of space of the region of interest
        v (Scalar): n-volume of the region of interest [nm^n]
        m (GenerationFunction): function of the random model used
        r (dict): parameters of the random model
        t (str): dislocations type
        p (VectorList): dislocation positions [nm]
        b (ScalarList): dislocation Burgers vectors sense [1]
        d (Scalar): density of dislocations [nm^-n]
        i (Scalar): inter dislocation distance [nm]
        c (str|None): conditions at the boundaries
    """

    geometries = ('square', 'circle') # available geometries

    @beartype
    def __init__(self,
        g: str,
        s: Scalar,
        m: GenerationFunction,
        r: dict,
        t: str = 'screw',
        c: Optional[str] = None,
    ) -> None:
        """
        Initialize the distribution.

        Input:
            g: geometry of the region of interest
            s: size of the region of interest [nm]
            m: model generating function
            r: parameters of the random model
            t: dislocations type
            c: conditions at the boundaries

        Complexity:
            O( complexity(m) )
        """
        # shape size
        if s <= 0:
            raise ValueError("incorrect size: "+str(s))
        self.s = s # characteristic size of the region of interest [nm]
        # shape geometry
        if not g in Distribution.geometries:
            raise ValueError("unknown geometry: "+str(g))
        self.g = g # geometry of the region of interest
        # shape n-volume
        if self.g == 'circle': # centered at the origin
            self.n = 2 # dimension of space
            self.v = np.pi*self.s**2 # surface [nm^2]
        elif self.g == 'square': # bottom left corner at the origin
            self.n = 2 # dimension of space
            self.v = self.s**2 # surface [nm^2]
        # model
        self.m = m # model generating function
        self.r = r.copy() # model parameters
        # dislocations
        self.t = t # dislocation type
        self.p, self.b = self.m(self.g, self.s, self.v, self.r) # generate
        self.d = len(self)/self.v # density of dislocations [nm^-n]
        self.i = 1/np.sqrt(self.d) # inter dislocation distance [nm]
        # boundary conditions
        self.c = c # boundary conditions code
        if not (self.c is None or c[:4]=='PBCR' and self.g=='square'):
            if self.g=='circle' and c=='IDBC':
                cp, cb = boundaries.IDBC(self.s, self.p, self.b, self.t)
            elif self.g=='square' and c[:4]=='PBCG':
                cp, cb = boundaries.PBCG(self.s, self.p, self.b, int(c[4:]))
            else:
                raise Exception("invalid boundary conditions: "+str(c))
            self.p = np.concatenate((self.p, cp))
            self.b = np.concatenate((self.b, cb))

    @beartype
    def __repr__(self) -> str:
        """
        Return the representation of the distribution.

        Output:
            r: string that can be evaluated
        """
        args = [
            repr(self.g), # geometry
            str(self.s), # size
            self.m.__name__, # model generating function
            repr(self.r), # model parameters
            repr(self.t), # dislocations type
            repr(self.c), # bondary conditions
        ]
        return "Distribution("+", ".join(args)+")"

    @beartype
    def __str__(self) -> str:
        """
        Return a textual description of the distribution.

        Output:
            s: distribution information
        """
        s = (
            "Distribution: "+self.fileName()
            + "\n- geometry: "+self.g
            + "\n- size: "+notation.number(self.s, 'console')+" nm"
            + "\n- n-volume: "+notation.number(self.v, 'console')
            + " nm^"+str(self.n)
            + "\n- dislocations type: "+self.t
            + "\n- model: "+self.m.__name__
            + notation.parameters(self.r, 'console')
            + "\n- population: "+str(len(self))+" dislocations"
            + "\n- dislocation density: "
            + notation.number(self.d*1e9**self.n, 'console')
            + " m^-"+str(self.n)
            + "\n- inter dislocation distance: "+str(round(self.i))+" nm"
            + "\n- boundary conditions: "+str(self.c)
            + "\n- b+: "+str(len(self.b[self.b>0]))+" dislocations"
            + "\n- b-: "+str(len(self.b[self.b<0]))+" dislocations"
        )
        return s

    @beartype
    def __len__(self) -> int:
        """
        Return the number of dislocations in the distribution.

        Output:
            n: number of Burgers vectors
        """
        return len(self.b)

    @beartype
    def fileName(self,
        t: bool = True,
    ) -> str:
        """
        Return a name that can be used to export the distribution.

        Input:
            t: add dislocation type information

        Output:
            n: file name of the distribution
        """
        n = (
            notation.number(self.d*1e9**self.n)+"m-"+str(self.n) # density
            + "_"+self.g+"_"+notation.number(self.s)+"nm" # geometry
            + "_"+self.m.__name__+notation.parameters(self.r) # model
        )
        if t:
            n += "_"+self.t # add dislocation type information
        if self.c:
            n += "_"+self.c # add bondary conditions information
        return n

    @beartype
    def plotTitle(self) -> str:
        """
        Return a name that can be used as a title in a plot.

        Output:
            t: title containing LaTeX code
        """
        m = self.m.__name__+notation.parameters(self.r, 'title') # model
        d = notation.number(self.d*1e9**self.n, 'title') # density
        t = m+r" $ \rho \sim "+d+r" m^{-2} $" # title
        if self.c and self.c[:4]!='PBCR':
            t += " "+self.c # boundary conditions
        return t

    @beartype
    def w(self,
        a: Vector,
        r: ScalarList,
        r2: ScalarList,
    ) -> ScalarList:
        """
        Return the edge correction coefficients of the shape.

        Input:
            a: position around which neighborhoods are formed [nm]
            r: radius of the neighborhoods [nm]
            r2: squared radius of neighborhoods [nm^2]

        Output:
            w: weighting coefficients for each value of radius around a

        Complexity:
            O( len(r) )
        """
        # vo: n-volumes of the overlappings (for each value of r)
        # vv: n-volumes of the vicinities of a (for each value of r)
        if self.g == 'circle':
            d2 = np.sum(np.square(a)) # squared distance to the origin of a
            d = np.sqrt(d2) # distance to the origin of a
            vo = overlap.circle_circle(r, self.s, d, r2, self.s**2, d2)
            vv = np.pi*r2
        elif self.g == 'square':
            vo = overlap.circle_square(a[0], a[1], r, r2, self.s)
            vv = np.pi*r2
        w = 1/np.divide(vo, vv, np.ones(vo.size), where=r2>0) # ratio
        return w

class Sample:
    """
    Represent a sample of distributions.

    Except for l, the attributes have the same meaning as for the
    Distribution class. The density and the inter dislocation distance
    are averaged over all distributions.

    Attributes:
        l (Tuple[Distribution, ...]): sampled distributions
        g (str): geometry of the region of interest
        s (Scalar): size of the region of interest [nm]
        n (int): dimension of space of the region of interest
        v (Scalar): n-volume of the region of interest [nm^n]
        m (GenerationFunction): function of the random model used
        r (dict): parameters of the random model
        t (str): dislocations type
        d (Scalar): averaged density of dislocations [nm^-n]
        i (Scalar): averaged inter dislocation distance [nm]
        c (str|None): conditions at the boundaries
    """

    @beartype
    def __init__(self,
        n: int,
        g: str,
        s: Scalar,
        m: GenerationFunction,
        r: dict,
        t: str = 'screw',
        c: Optional[str] = None,
    ) -> None:
        """
        Initialize the sample of distributions.

        Input:
            n: number of distributions to generate
            g: geometry of the region of interest
            s: size of the region of interest [nm]
            m: model generating function
            r: parameters of the random model
            t: dislocations type
            c: conditions at the boundaries

        Complexity:
            O( c * complexity(Distribution) )
        """
        if n <= 0:
            raise ValueError("incorrect number of distribution: "+str(n))
        self.l = tuple([Distribution(g, s, m, r, t, c) for i in range(n)])
        self.g = g # geometry of the region of interest
        self.s = s # size of the region of interest
        self.n = self[0].n # dimension of space
        self.v = self[0].v # n-volume of the region of interest
        self.m = m # model function
        self.r = r # model parameters
        self.t = t # dislocation type
        self.d = self.average(lambda d : d.d) # averaged dislocation density
        self.i = self.average(lambda d : d.i) # averaged inter disl. dist.
        self.c = c # boundary conditions

    @beartype
    def __repr__(self) -> str:
        """
        Return the representation of the sample of distributions.

        Output:
            r: string that can be evaluated
        """
        args = [
            str(len(self)), # number of distributions
            repr(self.g), # geometry
            str(self.s), # size
            self.m.__name__, # model function
            repr(self.r), # model parameters
            repr(self.t), # dislocation type
            repr(self.c), # boundary conditions
        ]
        return "Sample("+", ".join(args)+")"

    @beartype
    def __str__(self) -> str:
        """
        Return a textual description of the distribution.

        Output:
            s: distribution information
        """
        s = (
            "Sample: "+self.fileName()
            + "\n- population: "+str(len(self))+" distributions"
            + "\n- geometry: "+self.g
            + "\n- size: "+notation.number(self.s, 'console')+" nm"
            + "\n- n-volume: "+notation.number(self.v, 'console')
            + " nm^"+str(self.n)
            + "\n- dislocations type: "+self.t
            + "\n- model: "+self.m.__name__
            + notation.parameters(self.r, 'console')
            + "\n- dislocation density: "
            + notation.number(self.d*1e9**self.n, 'console')
            + " m^-"+str(self.n)
            + "\n- inter dislocation distance: "+str(round(self.i))+" nm"
            + "\n- boundary conditions: "+str(self.c)
        )
        return s

    @beartype
    def __getitem__(self, k: int) -> Distribution:
        """
        Return the k-th distribution stored in the sample.

        Input:
            k: index of the distribution

        Output:
            d: k-th distribution
        """
        return self.l[k]

    @beartype
    def __len__(self) -> int:
        """
        Return the number of distributions stored in the sample.

        Output:
            n: number of stored distributions
        """
        return len(self.l)

    @beartype
    def fileName(self, **kwargs) -> str:
        """
        Return a name that can be used to export the sample.

        The keyword arguments are passed to the equivalent function in
        the Distribution class.

        Output:
            n: file name of the samplee
        """
        return str(len(self))+'_'+self[0].fileName(**kwargs)

    @beartype
    def plotTitle(self) -> str:
        """
        Return a name that can be used as a title in a plot.

        Output:
            t: title containing LaTeX code
        """
        return str(len(self))+" "+self[0].plotTitle()

    @beartype
    def average(self,
        f: Callable[[Distribution, Any], AnalysisOutput],
        *args,
    ) -> AnalysisOutput:
        """
        Average the function f over the sample of distributions.

        The first parameter of f must be the distribution to analyze.
        The function f must return a summable object or a tuple of
        summable objects. In the case of a tuple the items are averaged
        one by one according to their position in the tuple.

        Input:
            f: function that can be applied to a distribution
            *args: additional arguments to pass to function f

        Outpur:
            r: result of f applied to each distribution and averaged

        Complexity:
            O( len(self) * complexity(f) )
        """
        r = f(self[0], *args) # result of f on the first distribution
        if isinstance(r, tuple):
            r = list(r)
            n = len(r)
            for i in range(1, len(self)):
                ri = f(self[i], *args) # result of f on the ith distribution
                for j in range(n):
                    r[j] += ri[j]
            for j in range(n):
                r[j] = r[j]/len(self)
            return tuple(r)
        else:
            for i in range(1, len(self)):
                r += f(self[i], *args)
            return r/len(self)



#!/usr/bin/env python
# coding: utf-8

"""
Tools for the generation of special conditions at the boundaries.
"""

from . import *

@beartype
def IDBC(
    s: Scalar,
    p: VectorList,
    b: ScalarList,
    t: str,
) -> Tuple[VectorList, ScalarList]:
    """
    Return the image dislocations of (p, b) for a circle of radius s.

    Adding the image dislocations allows to simulate correctly the
    complete diffraction profile (and not only for the asymptotic
    values of the Fourier variable). However, the formula used here
    only applies to dislocations of type 'screw'.

    Input:
        s: radius of the region of interest [nm]
        p: dislocation positions [nm]
        b: dislocation Burgers vectors sense [1]
        t: dislocation type

    Output:
        cp: image dislocation positions [nm]
        cb: image dislocation Burgers vectors sense [1]

    Complexity:
        O( len(p) )
    """
    if t == 'screw':
        n = np.linalg.norm(p, axis=1) # distance to the origin
        m = n != 0 # mask to avoid division by zero
        v = p[m] # points different from (0, 0)
        t = np.arctan2(v[:,1], v[:,0]) # angles of the image dislocations
        r = s**2/n[m] # radii of the image dislocations
        cp = np.stack((r*np.cos(t), r*np.sin(t)), axis=1) # image positions
        cb = -b # image Burgers vector senses
        return cp, -cb
    else:
        raise ValueError("cannot calculate images of edge dislocations")

@beartype
def PBCG(
    s: Scalar,
    p: VectorList,
    b: ScalarList,
    r: int,
) -> Tuple[VectorList, ScalarList]:
    """
    Return the replicated dislocations of (p, b) for a square of side s.

    Input:
        s: side of the region of interest [nm]
        p: dislocation positions [nm]
        b: dislocation Burgers vectors sense [1]
        r: number of replications at the boundaries

    Output:
        cp: replicated dislocation positions [nm]
        cb: replicated dislocation Burgers vectors sense [1]

    Complexity:
        O( len(p) * r^2 )
    """
    u = [] # list containing translated positions
    for i in range(1, r+1):
        for j in range(2*i):
            for k in (1, -1):
                u.append(np.array((-i*k, (i-j)*k)))
                u.append(np.array(((i-j)*k,  i*k)))
    cp = np.concatenate([p + s*d for d in u]) # replicated positions
    cb = np.tile(b, len(u)) # replicated Burgers vector senses
    return cp, cb



#!/usr/bin/env python
# coding: utf-8

"""
Tools for calculating the overlapping of two geometric objects.
"""

from . import *

@beartype
def circle_circle(
    rA: Union[Scalar, ScalarList],
    rB: Union[Scalar, ScalarList],
    d: Union[Scalar, ScalarList],
    r2A: Union[Scalar, ScalarList],
    r2B: Union[Scalar, ScalarList],
    d2: Union[Scalar, ScalarList],
) -> Union[Scalar, ScalarList]:
    """
    Return the overlapping area of two circles.

    Input:
        rA: radius of the circle A
        rB: radius of the circle B
        d: distance between the circle centers
        r2A: squared radius of the circle A
        r2B: squared radius of the circle B
        d2: squared distance between the circle centers

    Complexity:
        O( max(rA.size, rB.size, d.size) )
    """
    m0 = rA + rB <= d # mask: zero intersection
    mA = d + rA <= rB # mask: A is inside B
    mB = d + rB <= rA # mask: B is inside A
    m = np.logical_not(m0) & np.logical_not(mA) & np.logical_not(mB)
    # m is true for non-trivial cases
    o = np.subtract(
        np.add(
            np.multiply(
                r2A,
                np.arccos(
                    np.divide(
                        np.add(d2, np.subtract(r2A, r2B, where=m), where=m),
                        np.multiply(np.multiply(2, d, where=m), rA, where=m),
                        where=m),
                    where=m),
                where=m),
            np.multiply(
                r2B,
                np.arccos(
                    np.divide(
                        np.add(d2, np.subtract(r2B, r2A, where=m), where=m),
                        np.multiply(np.multiply(2, d, where=m), rB, where=m),
                        where=m),
                    where=m),
                where=m),
            where=m),
        np.divide(
            np.sqrt(
                np.multiply(
                    np.multiply(
                        np.subtract(np.add(rA, rB, where=m), d, where=m),
                        np.add(np.add(rA, rB, where=m), d, where=m),
                        where=m),
                    np.multiply(
                        np.add(np.subtract(rA, rB, where=m), d, where=m),
                        np.add(np.subtract(rB, rA, where=m), d, where=m),
                        where=m),
                    where=m),
                where=m),
            2,
            where=m),
        where=m)
    return np.where(mA, np.pi*r2A, np.where(mB, np.pi*r2B, o))

@beartype
def circle_square(
    x: Union[Scalar, ScalarList],
    y: Union[Scalar, ScalarList],
    r: Union[Scalar, ScalarList],
    r2: Union[Scalar, ScalarList],
    s: Union[Scalar, ScalarList],
) -> Union[float, ScalarList]:
    """
    Return the overlapping area of a circle and a square.

    Input:
        x: x coordinate of the circle center
        y: y coordinate of the circle center
        r: circle radius
        r2: squared circle radius
        s: square side

    Complexity:
        O( max(x.size, r.size, s.size) )
    """
    e = [] # list of areas of the circle outside the square for each frame
    dA = s - x # width of the right quadrant
    dB = s - y # height of the upper quadrant
    dC = x # width of the left quadrant
    dD = y # height of the lower quadrant
    # loop on the four quadrants and add their contributions to E
    for d1, d2 in [[dA, dB], [dB, dC], [dC, dD], [dD, dA]]:
        mq = d1**2 + d2**2 > r2
        m1 = mq & (d1 < r)
        m2 = mq & (d2 < r)
        mr = np.logical_not(mq)
        # area of the circle outside the square in the current quadrant
        eq = np.where(mr, np.subtract(np.pi*r2/4, d1*d2, where=mr), 0)
        for a in [[m1, d1], [m2, d2]]:
            np.add(
                np.divide(
                    np.subtract(
                        np.multiply(
                            r2,
                            np.arccos(
                                np.divide(a[1], r, where=a[0]),
                                where=a[0]),
                            where=a[0]),
                        np.multiply(
                            a[1],
                            np.sqrt(
                                np.subtract(
                                    r2,
                                    np.multiply(a[1], a[1], where=a[0]),
                                    where=a[0]),
                                where=a[0]),
                            where=a[0]),
                        where=a[0]),
                    2,
                    where=a[0]),
                eq,
                eq,
                where=a[0])
        e.append(eq)
    return np.pi*r2 - sum(e)



#!/usr/bin/env python
# coding: utf-8

"""
Standardized files for input to an X-ray diffraction simulation program.
"""

import os
from . import *
from . import sets

@beartype
def indices(
    v: Vector
) -> str:
    """
    Return the formated string of Miller indices of v.

    Input:
        v: Miller indices

    Output:
        s: formated string of Miller indices
    """
    sep, fmt = " ", '2.0f'
    return sep.join([format(c, fmt) for c in v])

@beartype
def contrast_factor(
    t: str,
    g: Vector,
    l: Vector,
    b: Vector,
    nu: Scalar,
) -> Scalar:
    """
    Return the contrast factor in elastically isotropic crystals.

    Input:
        t: dislocation type
        g: diffraction vector direction (hkl)
        l: dislocation line vector direction [uvw]
        b: Burgers vector direction [uvw]
        nu: Poisson's number [1]

    Output:
        C: dislocation contrast factor [1]
    """
    ng = np.linalg.norm(g) # diffraction vector norm to normalize
    nl = np.linalg.norm(l) # dislocation line vector norm to normalize
    psi = np.arccos(np.dot(g, l)/(ng*nl)) # angle between g and l
    if t == 'screw':
        C = np.sin(psi)**2 * np.cos(psi)**2
    elif t == 'edge':
        pg = g - (np.dot(g, l)/nl**2)*l # projection of g
        pb = b - (np.dot(b, l)/nl**2)*l # projection of b
        npg = np.linalg.norm(pg) # norm of the projection of g
        npb = np.linalg.norm(pb) # norm of the projection of b
        gamma = np.arccos(np.dot(pg, pb)/(npg*npb)) # angle betw. projections
        C = (np.sin(psi)**4
            / (8*(1-nu)**2)
            * (1-4*nu+8*nu**2+4*(1-2*nu)*np.cos(gamma)**2))
    return C

@beartype
def export_distribution(
    d: sets.Distribution,
    i: Scalar,
    p: str = "",
    n: Optional[str] = None,
    g: Vector = np.array([2, 0, 0]),
    b: Vector = np.array([1, 1, 0]),
) -> None:
    """
    Export the dislocations of d to a standardized input data file.

    The inter-dislocation distance must be specified because it may be
    different from d.i when averaged over several distributions.

    Input:
        d: distribution to be exported
        i: inter dislocation distance [nm]
        p: path where to place the exported file
        n: name of the exported file
        g: diffraction vector direction (hkl)
        b: Burgers vector direction [uvw]

    Complexity:
        O( len(d) )
    """
    if p!="" and p[-1]!="/":
        p += "/"
    if n == None:
        n = d.fileName()
    # parameters
    m = 12 # number of points along Lx
    a = 0.40494 # cell side [nm]
    a3 = max(2, i/m) # step size along Lx [nm]
    if d.g == 'circle':
        a3 = min(a3, 3*i/m)
    if d.t == 'screw':
        l = np.array([ 1,  1,  0]) # dislocation line vector direction [uvw]
        L = np.array([-1,  1,  0]) # direction along Lx [uvw]
        if np.linalg.norm(np.cross(l, b)) != 0:
            raise ValueError("screw type but l and b not parallel")
    elif d.t == 'edge':
        l = np.array([ 1, -1, -2]) # dislocation line vector direction [uvw]
        L = np.array([ 1,  1,  0]) # direction along Lx [uvw]
        if np.dot(l, b) != 0:
            raise ValueError("edge type but l and b not perpendicular")
    nu = 0.345 # Poisson's number
    C = contrast_factor(d.t, g, l, b, nu)
    if d.g == 'circle':
        str_g = "Cylinder radius"
    if d.g == 'square':
        if d.c and d.c[:4]=='PBCR':
            str_g = "Square_"+d.c[4:]+" side"
        else:
            str_g = "Square side"
    # write
    with open(p+n+".dat", "w") as f:
        h = ("# please keep the structure of this file unchanged\n"
            + indices(l)+" # z: direction of 'l' (line vector) [uvw]\n"
            + indices(L)+" # x: direction of 'L' (Fourier variable) [uvw]\n"
            + indices(b)+" # b: Burgers vector direction [uvw]\n"
            + indices(g)+" # g: diffraction vector direction (hkl)\n"
            + format(C,      '8.6f')+" # C: contrast coefficient [1]\n"
            + format(a,      '8.6f')+" # a: cell parameter [nm]\n"
            + format(d.s,    '8.0f')+" # s: "+str_g+" [nm]\n"
            + format(a3,     '8.1f')+" # a3: step size of 'L' along x [nm]\n"
            + format(nu,     '8.3f')+" # nu: Poisson's number [1]\n"
            + format(len(d), '8.0f')+" # number of dislocations\n"
            + "# Burgers vector and dislocation (x,y) coordinates\n")
        f.write(h)
        fmt = "%2.0f %27.20E %27.20E"
        np.savetxt(f, np.stack((d.b, d.p[:,0], d.p[:,1])).T, fmt=fmt)

@beartype
def export_sample(
    s: sets.Sample,
    p: str = "",
    n: Optional[str] = None,
) -> None:
    """
    Export a standardized input data file for each distributions of s.

    Input:
        s: sample of distributions to be exported
        p: path where to place the exported file
        n: name of the exported directory

    Complexity:
        O( len(s) * complexity(export_distribution) )
    """
    if p!="" and p[-1]!="/":
        p += "/"
    e = p+s.fileName()+"/" # folder where to export the distributions
    # export
    w = len(str(len(s))) # number of characters in file names
    if os.path.exists(e):
        for f in os.listdir(e):
            os.remove(e+f) # delete the existing distributions
    else:
        os.mkdir(e)
    for i in range(len(s)):
        export_distribution(s[i], s.i, e, str(i+1).zfill(w))

@beartype
def export(
    o: Union[sets.Distribution, sets.Sample],
    p: str = "",
    n: Optional[str] = None,
) -> None:
    """
    Convenient function for exporting a standardized data file.

    Input:
        o: distribution or sample of distributions to export
        p: path where to place the exported file
        n: name of the exported file or directory

    Complexity:
        O( complexity(export_distribution) ) if o is a distribution
        O( complexity(export_sample) ) if o is a sample
    """
    if isinstance(o, sets.Distribution):
        export_distribution(o, o.i, p, n)
    else:
        export_sample(o, p, n)



#!/usr/bin/env python
# coding: utf-8

"""
Tools for exporting the dislocation map of a distribution.
"""

import matplotlib.pyplot as plt
from . import *
from . import sets
from . import models

@beartype
def export(
    d: sets.Distribution,
    p: str = "",
    n: Optional[str] = None,
    t: Optional[str] = None,
) -> None:
    """
    Export the dislocation map of the distribution d.

    Input:
        d: distribution to be exported
        p: path where to export the file
        n: name of the exported file
        t: title of the plot

    Complexity:
        O( len(d) )
    """
    if p!="" and p[-1]!="/":
        p += "/"
    if n == None:
        n = d.fileName(t=False)
    if t == None:
        t = d.plotTitle()
    fig, ax = plt.subplots(figsize=(6, 6))
    # aspect
    ax.set_aspect(1)
    b = d.s * 0.05 # borders width
    s, w = 100, 0.2 # marker size and line width
    # grid
    if d.m in (models.RRDD, models.RRDD) and (d.c is None or d.c[:4]=='PBCR'):
        ax.grid(True, zorder=0) # subareas or cells grid
        ticks = models.ticks(d.g, d.s, d.r['s'])
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        if len(ticks) > 10: # the grid is too thin to display all the ticks
            labels = ["" for i in range(len(ticks))]
            labels[0], labels[-1] = round(ticks[0]), round(ticks[-1])
            ax.set_xticklabels(labels)
            ax.set_yticklabels(labels)
    # region of interest
    if d.g == 'circle':
        g = plt.Circle((0,0), d.s, color='k', fill=False, zorder=50)
        if d.c is None:
            ax.set_xlim([-d.s-b, d.s+b])
            ax.set_ylim([-d.s-b, d.s+b])
        elif d.c == 'IDBC':
            r = 1.5
            ax.set_xlim([-b-(r+0.5)*2*d.s, b+(r+0.5)*2*d.s])
            ax.set_ylim([-b-(r+0.5)*2*d.s, b+(r+0.5)*2*d.s])
            k = 2*r + 1
            w /= k
            s /= k**2
    else:
        g = plt.Rectangle((0,0), d.s, d.s, color='k', fill=False, zorder=50)
        if d.c is None:
            ax.set_xlim([-b, d.s+b])
            ax.set_ylim([-b, d.s+b])
        elif d.c[:4]=='PBCG':
            r = int(d.c[4:])
            ax.set_xlim([-b-r*d.s, d.s+b+r*d.s])
            ax.set_ylim([-b-r*d.s, d.s+b+r*d.s])
            k = 2*r + 1
            w /= k
            s /= k**2
    ax.add_artist(g)
    # dislocations
    partition = [ # positive and negative Burgers vector senses
        (r"$ \bot $", "Burgers vector sense $+$", d.p[d.b>0]),
        (r"$ \top $", "Burgers vector sense $-$", d.p[d.b<0]),
    ]
    for k in partition:
        x, y = k[2][:,0], k[2][:,1]
        plt.scatter(
            x,
            y,
            marker=k[0],
            label=k[1],
            zorder=100,
            linewidths=w,
            s=s,
        )
    # information
    ax.set_xlabel(r"$x \ (nm)$")
    ax.set_ylabel(r"$y \ (nm)$")
    plt.title(t)
    l = plt.legend(facecolor='white', framealpha=1)
    l.set_zorder(150)
    # export
    plt.savefig(p+n+".pdf")
    plt.close('all')



#!/usr/bin/env python
# coding: utf-8

"""
Tools for parallelizing the spatial analysis of distributions.
"""

from mpi4py import MPI
name = MPI.Get_processor_name()
comm = MPI.COMM_WORLD
rank = comm.Get_rank() # number of the processor executing this script
size = comm.Get_size() # number of processors
root = 0 # master processor

from . import *
from . import sets
from . import analyze

def average_on_cores(
    w,
    b: bool = False
):
    """
    Return the average value of w ​​over the cores.

    When b is True, all cores get the average value. When b is False,
    the workers get None and the root gets the average value.

    Input:
        w: worker value
        b: broadcast the result to all cores

    Output:
        m: averaged value of w over the cores
    """
    if not isinstance(w, np.ndarray):
        w = np.array(w)
    if rank == root: # if the script is executed by the master
        m = np.zeros_like(w) # prepare the buffer for data reception
    else: # if the script is executed by a worker
        m = None
    comm.Reduce([w, MPI.DOUBLE], [m, MPI.DOUBLE], op=MPI.SUM, root=root)
    if rank == root:
        m = m/size # average the value
    if b:
        m = comm.bcast(m, root=root) # broadcast to the workers
    return m

def export(
    o: Union[sets.Distribution, sets.Sample],
    p: str = "",
    n: Optional[str] = None,
    t: Optional[str] = None,
) -> None:
    """
    Export a complete pooled analysis of the object o of each core.

    Function similar to the export function of the analyze module.

    Input:
        o: distribution or sample to analyze on the core
        p: path where to export the file
        n: name of the exported file
        t: title of the plot
    """
    if p!="" and p[-1]!="/":
        p += "/"
    i = average_on_cores(o.i, True) # averaged inter dislocation distance
    r, iK = analyze.intervals(i, o.s) # intervals to display
    f = ['KKKK', 'gggg', 'GaGs'] # functions to calculate
    worker = analyze.calculate(f, o, r) # function results
    master = [average_on_cores(worker[i]) for i in range(len(worker))]
    if rank == root:
        if isinstance(o, sets.Distribution):
            c = str(size) # number of distributions analyzed
            t = c+" "+o.plotTitle() # plots title
            n = c+"_"+o.fileName() # plots file name
        else:
            c = str(len(o)*size) # number of distributions analyzed
            t = c+" "+o[0].plotTitle() # plots title
            n = c+"_"+o[0].fileName() # plots file name
        analyze.plot_KKKK(r[:iK], master[f.index('KKKK')].T[:iK].T, p, n, t)
        analyze.plot_gggg(r, master[f.index('gggg')], p, n, t)
        analyze.plot_GaGs(r, master[f.index('GaGs')], p, n, t)
